#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>

#include "wrappers.h"

/* Naming conventions: 
 * x* = exit on failure
 * r* = retry
 * w* = warn on failure 
 */

int xsocket(int domain, int type, int protocol) {
	int socketfd = socket(domain, type, protocol);
	if (socketfd == -1) show_error_and_exit();
	return socketfd;
}

void xbind(int sockfd, struct sockaddr *my_addr, size_t addrlen) {
	int retval = bind(sockfd, my_addr, addrlen);
	if (retval == -1) show_error_and_exit();
}

void xlisten(int sockfd, int backlog) {
	int retval = listen(sockfd, backlog);
	if (retval == -1) show_error_and_exit();
}

void xconnect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen) {
	int retval = connect(sockfd, serv_addr, addrlen);
	if (retval == -1) show_error_and_exit();
}

/* Given an ip address (not a hostname) and a port, try to connect to it, and
   return the filehandle */
int connect_to_host_and_port(char *host_ip, int port) {
        int socketfd;
        struct sockaddr_in si;

        si.sin_family = AF_INET;
        si.sin_port = htons(port);
        inet_aton(host_ip, &si.sin_addr);
        memset(&(si.sin_zero), '\0', 8);

        socketfd = socket(PF_INET, SOCK_STREAM, 0);
        if (socketfd < 0) {
		warn("Couldn't connect to %s %i\n", host_ip, port);
		return socketfd;
	}

        if (connect(socketfd, (const struct sockaddr *) &si, sizeof(si)) == -1) {
		warn("Connecting to %s:%i failed.\n", host_ip, port);
		return -1;
	}
	return socketfd;
}


#ifndef FORSATABS
/* Satabs (rightly enough) doesn't like *(&buf + bytes_recv) = '\0';
 * it does rely on the caller to get the size of the buffer right.
 * This function should be rewritten so satabs accepts it. */
/* accept an incoming connection, read one piece of data,
   and then CLOSE the connection.  
   Warning: the buffer will be null-terminated. 
   returns bytes received*/
ssize_t wget_data(int sockfd, char *buf, size_t size) {
	int connfd;
	ssize_t bytes_recv;
	socklen_t struct_size;
	struct sockaddr_in ri;
	
	struct_size = sizeof(ri);
	connfd = accept(sockfd, (struct sockaddr *) &ri, &struct_size);
	if (connfd == -1) {
		warn("Accept failed.\n");
		return -1;
	}
	if (debug) {
		warn("received a conn\n");	
	}
	/* Assume that everything comes in one recv.  This is a 
	faulty assumption, but it works well enough for this demo.*/
	bytes_recv = recv(connfd, &buf, size, 0);
	if (bytes_recv <= 0) {
		warn("Recieving data failed. (%s)\n", describe_err(errno));
		wclose(connfd);
		return bytes_recv;
	}
	
	/* The following condition should never be true, 
	but satabs doesn't know this. */
	if (bytes_recv >= size) {
		bytes_recv = size - 1;
	}
	*(&buf + bytes_recv) = '\0'; 
	wclose(connfd);
	return bytes_recv;
}
#endif /*FORSATABS*/

/* refactor wget_recv to use this? 
   This does _not_ close the connection on failure
*/
char *wrecv_nullterminated(int connfd, int buf_size) {
	ssize_t bytes_recv;
	//char *buf = malloc(buf_size);
	char buf[buf_size];

	//if (buf == NULL) return NULL;
	memset(buf, '\0', buf_size);	
	/* the -1 guarantees that it stays null-terminated */
	bytes_recv = recv(connfd, &buf, buf_size -1, 0);
	if (bytes_recv <= 0) {
		warn("Recieving data failed. (%s)\n", describe_err(errno));
		//free(buf);
                return NULL;
        }
	
	return strdup(buf);
}

#ifndef FORSATABS /* hide this: satabs can't deal with the recursion */
/* wrapper around SEND that handles EINTR: 'retry send' 
 * If any other errors occur, -1 is returned with errno set
 * as for normal send */
ssize_t rsend(int s, const void *buf, size_t len, int flags) {
	ssize_t sent = send(s, buf, len, flags);
	if (sent == -1) {
		if (errno == EINTR) {
		/* yes, this uses more stack than needed... 
		 * this shouldn't be a serious issue */
			rsend(s, buf, len, flags);
		} else {
			warn("Sending failed.\n");
		}
	}
	
	return sent;
}
#endif /* FORSATABS */

void show_error_and_exit() {
	printf("Quitting.  %s\n", describe_err(errno));
	exit(2);
}


int wclose(int fd) {
	int retval = close(fd);
	if (retval == -1) {
		warn("Warning: closing fd %d failed. %s\n", describe_err(errno));
	}
	return retval;
}


/* strerror wrapper which preserves errno */
char *describe_err(int err) {
	int save_err = errno;
	char *desc = strerror(err);
	errno = save_err;
	return desc;
}

/* thin wrapper around printf which preserves errno and only prints when
 * in debug mode */
#ifndef FORSATABS /* hack: satabs handles varargs wrong if the body is defined */
void warn(const char *fmt, ...) {
	if (debug) {
		int tmperr = errno;
		va_list argp;
		va_start(argp, fmt);
		vprintf(fmt, argp);
		va_end(argp);
		errno = tmperr;
	}
	
}
#endif /* FORSATABS */

/* this is a workaround, as strjoin uses varargs, which the old satabs parser
   doesn't deal with nicely, and so the body of that is preprocessed out when
   we're verifying, so satabs doesn't realize that it dynamically allocates
   memory, and so dies when that memory is freed in the caller.
*/
void free_after_dyn_not_recognized_by_satabs(void *a_ptr) {
	#ifndef FORSATABS
	free(a_ptr);
	#endif
}

