/* The location net is an abstract concept.  The exact form of the
   location ID can change freely, from ip addresses to hostnames to
   location IDs which are unique within an example, such as "a" and "b".
   
   In this implementation, a tree -will be- provided which contains
   mappings between locations and underlying IPv4 addresses.  
  
   A server is queried for the run-time location net information.  
*/

/* common location functions: these are shared between implementations of
   location.c (see location.c for details on why there are multiple
   implementations.) */

#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* memset */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "wrappers.h"
#include "cstring.h"
#include "clocation.h"

#define LOCATION_LOOKUP_SERVER_HOST "127.0.0.1"
#define LOCATION_LOOKUP_SERVER_PORT 2327

static char* lookup_location_ip(char *location_name);
static int lookup_location_port(char *location_name);
static void clean_response(char *response);
static char *make_location_ip_lookup(char *location_name);
static void clean_response(char *response);
static char *make_location_port_lookup(char *location_name);


/* Hide the visibility of these function from external source files. */

static char* do_lookup(char *request) {
	int socketfd;
	char *response;
	
	if (request == NULL) return NULL;
	socketfd = connect_to_host_and_port(LOCATION_LOOKUP_SERVER_HOST,
						LOCATION_LOOKUP_SERVER_PORT);
	warn("socketfd: %i\n", socketfd);
	if (socketfd == -1) return NULL;
	if (rsend(socketfd, request, strlen(request), 0) <= 0) return NULL;
	warn("rsend done (lli)\n");
	response = wrecv_nullterminated(socketfd, MAX_LOCATION_LEN);
	warn("wn done (lli)\n");
	wclose(socketfd);
	clean_response(response);
	
	return response;	
}

char* lookup_location_ip(char *location_name) {
	return do_lookup(make_location_ip_lookup(location_name));
}

/* this returns 0 on failure. */
static int lookup_location_port(char *location_name) {
	int port;
	char *raw_port = do_lookup(make_location_port_lookup(location_name));
	if (raw_port == NULL) return 0;
	port = strtol(raw_port, NULL, 10);
	free_after_dyn_not_recognized_by_satabs(raw_port);
	return port;
}

/* replace the trailing newline with a '\0' */
static void clean_response(char *response) {
	#ifndef FORSATABS
	int index = 0;
	
	if (response == NULL) return;
	while (response[index] != '\0') {
		if (response[index] == '\n') {
			response[index] = '\0';
			return;
		}
		index++;
	}
	#endif /* FORSATABS */
}

static char *make_location_ip_lookup(char *location_name) {
	int buf_size = strlen(location_name) + 5;
	char *buf = malloc(buf_size);
	if (buf == NULL) return NULL;
	sprintf(buf, "ip: %s\n", location_name);
	return buf;
}

static char *make_location_port_lookup(char *location_name) {
	int buf_size = strlen(location_name) + 7;
        char *buf = malloc(buf_size);
        if (buf == NULL) return NULL;
        sprintf(buf, "port: %s\n", location_name);
        return buf;
}

/* 'public' API follows */

/* the following two setters aren't very meaningful except
   when running (as opposed to verifying) the code. */

int set_applet_origin(char *origin, char *full_path_to_applet) {
	char *envstring;
	envstring = strjoin(4, "ORIGIN", full_path_to_applet, "=", origin);

	if (envstring == NULL) {
		return -1;
	}
	warn("envstring is %s\n", envstring);
	putenv(envstring);
	free_after_dyn_not_recognized_by_satabs(envstring);

	return 0;
}

int set_inet_origin(char *host, char *port, char *req_name, char *local_name) {
	char *origin;
	int retval;
	origin = strjoin(4, host, ":", port, req_name);
	if (origin == NULL) {
		return -1;
	} 
	/* FIXME: make sure that local_name is a full path */
	retval =  set_applet_origin(origin, local_name);	
	free_after_dyn_not_recognized_by_satabs(origin);
	
	return retval;
}


/* to 'disconnect', for now, just use close, though this
   only closes one fd, rather than all which are connected
   to a location */
int connect_to_location(char *location_name) {
        char *ip = lookup_location_ip(location_name);
        int port = lookup_location_port(location_name);
        if ((ip == NULL) || (port <= 0)) {
                warn("Looking up location %s failed.\n", location_name);
                return -1;
        }
        return connect_to_host_and_port(ip, port);
}


