#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <stdio.h> /* FIXME: for snprintf - move stuff that needs it */

#ifndef FORSATABS
#include <fcntl.h>
#endif /* FORSATABS */

#include "wrappers.h"
#include "run.h"
#include "cstring.h"

/* filename must be <= MAX_FILENAME_LENGTH.  If filename is not absolute,
 * ./filename must also be <= MAX_FILENAME_LENGTH, including terminating NULL.
 * This truncation is unfortunate, but necessary for verification 
 */
int add_path_to_filename(char *filename, char *buffer, size_t buffer_size);
int is_absolute_filename(char *filename);
int grant_user_perms(char *filename);
int get_executable_name(char *filename, char *buffer, size_t buffer_size);
int is_source_agent(char *filename);
char* create_exec_agent(char *filename);
int untar_source_to_workdir(char *filename, char *workdir);
int get_work_dir_name(char *filename, char *buf, size_t buf_size);
char* get_agent_name(char *workdir);
int build_agent(char *filename, char *workdir);
int get_bin_name(char *workdir, char *buf, size_t buflen);
char *path_join(char *part0, char *part1);
int run_agent_with_args(char *filename, char **argv);

/* fork off a new process and run the agent in it */
int run_agent(char *filename) {
	char *argv[2];
	argv[0] = filename;
	argv[1] = NULL;
	return run_agent_with_args(filename, argv);
}

int run_agent_with_args(char *filename, char **argv) {
    char filename_buffer[MAX_FILENAME_LENGTH];
	pid_t cur_pid;
	
	if (grant_user_perms(filename) == -1) {
		warn("Uh-oh.. granting permissions on %s failed.  Continuing.\n", filename);
	}

	get_executable_name(filename, filename_buffer, MAX_FILENAME_LENGTH);

	cur_pid = fork();
	if (cur_pid < 0) {
		warn("fork failed. Agent will not be executed.");
		return cur_pid;
	} else if (cur_pid == 0) { /* child process */
		execv(filename_buffer, argv);
		/* if execv is successful, the following will not be reached */
		warn("If you see this, executing %s failed (%s).\n",
			filename_buffer, describe_err(errno));
		exit(2); /* this will not exit the parent ...*/
	} else { /* parent process */
		return 0;
	}
}

/* make sure that the file has read and execute permissions for the user,
 * so that checking whether it's a source or executable agent can be done,
 * and so that it can be run if it's an executable agent. 
 */
int grant_user_perms(char *filename) {
	struct stat st;
	mode_t current_perms;
	mode_t add_perms = S_IXUSR | S_IRUSR;
	if (stat(filename, &st) == -1) {
		warn("stat'ing %s failed.. not good.\n", filename);
		return chmod(filename, add_perms);
	}
	current_perms = st.st_mode;
	return chmod(filename, current_perms|add_perms);
}


/* Get the name (including either an absolute path, or a path starting
 * with ./) of the executable to run.  If this executable must be built,
 * it will be ... rename this function?
 * Documentation on what kinds of files are considered agents and how those
 * are treated can be found in Documentation/agents.txt
 */
int get_executable_name(char *filename, char *buffer, size_t buffer_size) {
	char *exec_filename;
	
	if (is_source_agent(filename)) {
		exec_filename = create_exec_agent(filename);
	} else {
		exec_filename = filename;
	}
	
	return add_path_to_filename(exec_filename, buffer, buffer_size);
}

#ifndef FORSATABS
int is_absolute_filename(char *filename) {
     /* If the filename is relative, include ./ at the start so that
     * it can be executed even if '.' isn't in the path.
     * The test for relativity is unix-centric...*/
     return filename[0] == '/';
}
#endif /* FORSATABS */


#define RESERVE_BYTES 3 /* The magic constant 3 is the length of './', plus NULL */
int add_path_to_filename(char *filename, char *filename_buffer, size_t buffer_size) {
    /* If the filename is relative, include ./ at the start so that
     * it can be executed even if '.' isn't in the path.
     * The test for relativity is unix-centric...*/
    int filename_len;
	
	if (filename == NULL) return -1;
	
	filename_len = strlen(filename);
    if (filename_len < 0) return -1; /* impossible case, here for SATABS */

    if (is_absolute_filename(filename)) {
        strncpy(filename_buffer, filename, buffer_size);
	if (strlen(filename) >= buffer_size) 
		filename_buffer[buffer_size - 1] = '\0';
    } else {
        if (filename_len > buffer_size - RESERVE_BYTES) {
            filename_len = buffer_size - RESERVE_BYTES;
        }

        filename_buffer[0] = '.';
        filename_buffer[1] = '/';
        strncpy(filename_buffer + 2, filename, filename_len);
		filename_buffer[filename_len + 2] = '\0';
    }
    return 0; /* success */
}

/*
		        exec_filename = create_exec_agent(filename);
*/

/* 1 if it is a source agent, 0 otherwise */
#define GZIP_HEADER_MAGIC0 0x1f
#define GZIP_HEADER_MAGIC1 0x8b
#define HEADER_BYTES_TO_CHECK 2
/* FIXME: define a wrapper that deals with EINTR */
int is_source_agent(char *filename) {
	int fd, bytes_read, is_source_agent;
	unsigned char header[HEADER_BYTES_TO_CHECK];
	
	fd = open(filename, O_RDONLY);
	if (fd == -1) {
		warn("Unable to open %s (%s)\n", filename, describe_err(errno));
		return 0;
	}
	
	bytes_read = read(fd, header, HEADER_BYTES_TO_CHECK);
	if (bytes_read == -1) {
		warn("Error reading %s (%s)\n", filename, describe_err(errno));
		wclose(fd);
		return 0;
	}

	if (bytes_read != HEADER_BYTES_TO_CHECK) {
		warn("Read less than %d bytes from %s; file is probably invalid\n",
			HEADER_BYTES_TO_CHECK, filename);
		wclose(fd);
		return 0;
	}
	wclose(fd);

	is_source_agent = 1;
	is_source_agent &= (header[0] == GZIP_HEADER_MAGIC0);
	is_source_agent &= (header[1] == GZIP_HEADER_MAGIC1);
	return is_source_agent;
	
}

/* Given the filename of a source agent, run its Makefile and
 * return the name of the binary produced.  The specification
 * of what should be in a source agent is provided in 
 * Documentation/agent.txt 
 */
char* create_exec_agent(char *filename) {
	char workdir[MAX_FILENAME_LENGTH];
	
	if (get_work_dir_name(filename, workdir, MAX_FILENAME_LENGTH) == -1) {
		warn("Couldn't come up with a build directory name for %s.\n", filename);
		return NULL;
	}
	
	if (mkdir(workdir, 0700) == -1) {
		warn("creating %s failed (%s)\n", workdir, describe_err(errno));
		return NULL;
	}
	
	if (build_agent(filename, workdir) == -1) {
		warn("building agent from %s failed.\n");
		return NULL;
	}
	
	return get_agent_name(workdir);
}

int build_agent(char *filename, char *workdir) {
	int cur_dir_fd;
	
	if ((cur_dir_fd = open(".", O_RDONLY)) == -1) {
		warn("Opening the current directory failed (%s).\n", describe_err(errno));
		return -1;
	}
	
	if (untar_source_to_workdir(filename, workdir) == -1) {
		warn("Untar of %s to %s failed. Not building agent.\n");
		wclose(cur_dir_fd);
		return -1;
	}
	
	if (chdir(workdir) == -1) {
		warn("Unable to change to %s (%s). Not building agent.\n", 
					workdir, describe_err(errno));
		wclose(cur_dir_fd);
		return -1;
	}

	if (system("make") == -1) {
		warn("Executing make for %s failed. Continuing.\n", filename);
	}

	if (fchdir(cur_dir_fd) == -1) {
		warn("Changing back to old current directory failed (%s).\n", 
					describe_err(errno));
	}
	
	wclose(cur_dir_fd);
	return 0;
}

/* FIXME: this is a unix-ism. Also, it will fail in interesting ways
 * if filename or workdir are interpreted by untar or the shell - ie, if
 * they contain ; in their names, start with -, etc. */
int untar_source_to_workdir(char *filename, char *workdir) {
	#define CSTW_CMD_BUF_SIZE MAX_FILENAME_LENGTH * 2 + 50
	char cmd_buf[CSTW_CMD_BUF_SIZE]; /* allocate a bit more space than needed */
	if (snprintf(cmd_buf, 
				 CSTW_CMD_BUF_SIZE, 
				 "tar -C %s -xvzf %s", 
				 workdir, 
				 filename) == -1) {
			return -1;
	}
	return system(cmd_buf);
}

/* return the name for the workdir - it's just the filename + "dir" at 
 * the end.  If this would be too long, the filename part is truncated.
 */
int get_work_dir_name(char *filename, char *buf, size_t buf_size) {
	int filename_len = strlen(filename);
	int dir_suffix_start = filename_len;
	char *dir_suffix = "dir";
	int space_needed_for_suffix = 4; /* strlen("dir") + 1 for NUL */
	
	if (filename == NULL) return -1;
	
	if (dir_suffix_start + space_needed_for_suffix > buf_size) {
		dir_suffix_start = buf_size - space_needed_for_suffix;
	}
	if (dir_suffix_start < 0) {
		/* buffer is smaller than the suffix */
		return -1;
	}
	

	strncpy(buf, filename, buf_size);
	strncpy(buf + dir_suffix_start, dir_suffix, space_needed_for_suffix);
	return 0;
}


char* get_agent_name(char *workdir) {
	char runbin_name[MAX_FILENAME_LENGTH];
	
	if (get_bin_name(workdir, runbin_name, MAX_FILENAME_LENGTH) == -1) {
		return NULL;
	}
	
	return path_join(workdir, runbin_name);
}


int get_bin_name(char *workdir, char *buf, size_t buflen) {
	char *runbin_name = path_join(workdir, "runbin");
	int fd;

	if (runbin_name == NULL) return -1;
	fd = open(runbin_name, O_RDONLY);
	free_after_dyn_not_recognized_by_satabs(runbin_name);
	if (fd == -1) {
		warn("Unable to open %s (%s).\n", runbin_name, describe_err(errno));
		return -1;
	}

	/* FIXME: handle EINTR */
	return read(fd, buf, buflen);
}

char *path_join(char *part0, char *part1) {
	return strjoin(3, part0, "/", part1);
}


